@@ -3,8 +3,8 @@ import argparse
 import collections
 import logging
 import socket
-import sys
-import threading
+# https://github.com/python/typeshed/blob/master/stdlib/2and3/socket.pyi
+from socket import errno as socket_errors  # type: ignore
 
 import OpenSSL
 import six
@@ -12,8 +12,10 @@ import zope.interface
 
 from acme import challenges
 from acme import standalone as acme_standalone
+# pylint: disable=unused-import, no-name-in-module
+from acme.magic_typing import DefaultDict, Dict, Set, Tuple, List, Type, TYPE_CHECKING
 
-from certbot import cli
+from certbot import achallenges  # pylint: disable=unused-import
 from certbot import errors
 from certbot import interfaces
 
@@ -21,6 +23,11 @@ from certbot.plugins import common
 
 logger = logging.getLogger(__name__)
 
+if TYPE_CHECKING:
+    ServedType = DefaultDict[
+        acme_standalone.BaseDualNetworkedServers,
+        Set[achallenges.KeyAuthorizationAnnotatedChallenge]
+    ]
 
 class ServerManager(object):
     """Standalone servers manager.
@@ -35,14 +42,12 @@ class ServerManager(object):
     will serve the same URLs!
 
     """
-    _Instance = collections.namedtuple("_Instance", "server thread")
-
     def __init__(self, certs, http_01_resources):
-        self._instances = {}
+        self._instances = {}  # type: Dict[int, acme_standalone.BaseDualNetworkedServers]
         self.certs = certs
         self.http_01_resources = http_01_resources
 
-    def run(self, port, challenge_type):
+    def run(self, port, challenge_type, listenaddr=""):
         """Run ACME server on specified ``port``.
 
         This method is idempotent, i.e. all calls with the same pair of
@@ -51,35 +56,35 @@ class ServerManager(object):
         :param int port: Port to run the server on.
         :param challenge_type: Subclass of `acme.challenges.Challenge`,
             either `acme.challenge.HTTP01` or `acme.challenges.TLSSNI01`.
+        :param str listenaddr: (optional) The address to listen on. Defaults to all addrs.
 
-        :returns: Server instance.
+        :returns: DualNetworkedServers instance.
         :rtype: ACMEServerMixin
 
         """
         assert challenge_type in (challenges.TLSSNI01, challenges.HTTP01)
         if port in self._instances:
-            return self._instances[port].server
+            return self._instances[port]
 
-        address = ("", port)
+        address = (listenaddr, port)
         try:
             if challenge_type is challenges.TLSSNI01:
-                server = acme_standalone.TLSSNI01Server(address, self.certs)
+                servers = acme_standalone.TLSSNI01DualNetworkedServers(
+                    address, self.certs)  # type: acme_standalone.BaseDualNetworkedServers
             else:  # challenges.HTTP01
-                server = acme_standalone.HTTP01Server(
+                servers = acme_standalone.HTTP01DualNetworkedServers(
                     address, self.http_01_resources)
         except socket.error as error:
             raise errors.StandaloneBindError(error, port)
 
-        thread = threading.Thread(
-            # pylint: disable=no-member
-            target=server.serve_forever)
-        thread.start()
+        servers.serve_forever()
 
         # if port == 0, then random free port on OS is taken
         # pylint: disable=no-member
-        real_port = server.socket.getsockname()[1]
-        self._instances[real_port] = self._Instance(server, thread)
-        return server
+        # both servers, if they exist, have the same port
+        real_port = servers.getsocknames()[0][1]
+        self._instances[real_port] = servers
+        return servers
 
     def stop(self, port):
         """Stop ACME server running on the specified ``port``.
@@ -88,13 +93,12 @@ class ServerManager(object):
 
         """
         instance = self._instances[port]
-        logger.debug("Stopping server at %s:%d...",
-                     *instance.server.socket.getsockname()[:2])
-        instance.server.shutdown()
+        for sockname in instance.getsocknames():
+            logger.debug("Stopping server at %s:%d...",
+                         *sockname[:2])
         # Not calling server_close causes problems when renewing multiple
         # certs with `certbot renew` using TLSSNI01 and PyOpenSSL 0.13
-        instance.server.server_close()
-        instance.thread.join()
+        instance.shutdown_and_server_close()
         del self._instances[port]
 
     def running(self):
@@ -103,50 +107,68 @@ class ServerManager(object):
         Once the server is stopped using `stop`, it will not be
         returned.
 
-        :returns: Mapping from ``port`` to ``server``.
+        :returns: Mapping from ``port`` to ``servers``.
         :rtype: tuple
 
         """
-        return dict((port, instance.server) for port, instance
-                    in six.iteritems(self._instances))
+        return self._instances.copy()
 
 
-SUPPORTED_CHALLENGES = [challenges.TLSSNI01, challenges.HTTP01]
+SUPPORTED_CHALLENGES = [challenges.TLSSNI01, challenges.HTTP01] \
+# type: List[Type[challenges.KeyAuthorizationChallenge]]
 
 
-def supported_challenges_validator(data):
-    """Supported challenges validator for the `argparse`.
+class SupportedChallengesAction(argparse.Action):
+    """Action class for parsing standalone_supported_challenges."""
 
-    It should be passed as `type` argument to `add_argument`.
+    def __call__(self, parser, namespace, values, option_string=None):
+        logger.warning(
+            "The standalone specific supported challenges flag is "
+            "deprecated. Please use the --preferred-challenges flag "
+            "instead.")
+        converted_values = self._convert_and_validate(values)
+        namespace.standalone_supported_challenges = converted_values
 
-    """
-    if cli.set_by_cli("standalone_supported_challenges"):
-        sys.stderr.write(
-            "WARNING: The standalone specific "
-            "supported challenges flag is deprecated.\n"
-            "Please use the --preferred-challenges flag instead.\n")
-    challs = data.split(",")
+    def _convert_and_validate(self, data):
+        """Validate the value of supported challenges provided by the user.
 
-    # tls-sni-01 was dvsni during private beta
-    if "dvsni" in challs:
-        logger.info("Updating legacy standalone_supported_challenges value")
-        challs = [challenges.TLSSNI01.typ if chall == "dvsni" else chall
-                  for chall in challs]
-        data = ",".join(challs)
+        References to "dvsni" are automatically converted to "tls-sni-01".
 
-    unrecognized = [name for name in challs
-                    if name not in challenges.Challenge.TYPES]
-    if unrecognized:
-        raise argparse.ArgumentTypeError(
-            "Unrecognized challenges: {0}".format(", ".join(unrecognized)))
+        :param str data: comma delimited list of challenge types
 
-    choices = set(chall.typ for chall in SUPPORTED_CHALLENGES)
-    if not set(challs).issubset(choices):
-        raise argparse.ArgumentTypeError(
-            "Plugin does not support the following (valid) "
-            "challenges: {0}".format(", ".join(set(challs) - choices)))
+        :returns: validated and converted list of challenge types
+        :rtype: str
 
-    return data
+        """
+        challs = data.split(",")
+
+        # tls-sni-01 was dvsni during private beta
+        if "dvsni" in challs:
+            logger.info(
+                "Updating legacy standalone_supported_challenges value")
+            challs = [challenges.TLSSNI01.typ if chall == "dvsni" else chall
+                      for chall in challs]
+            data = ",".join(challs)
+
+        unrecognized = [name for name in challs
+                        if name not in challenges.Challenge.TYPES]
+
+        # argparse.ArgumentErrors raised out of argparse.Action objects
+        # are caught by argparse which prints usage information and the
+        # error that occurred before calling sys.exit.
+        if unrecognized:
+            raise argparse.ArgumentError(
+                self,
+                "Unrecognized challenges: {0}".format(", ".join(unrecognized)))
+
+        choices = set(chall.typ for chall in SUPPORTED_CHALLENGES)
+        if not set(challs).issubset(choices):
+            raise argparse.ArgumentError(
+                self,
+                "Plugin does not support the following (valid) "
+                "challenges: {0}".format(", ".join(set(challs) - choices)))
+
+        return data
 
 
 @zope.interface.implementer(interfaces.IAuthenticator)
@@ -169,14 +191,15 @@ class Authenticator(common.Plugin):
         self.key = OpenSSL.crypto.PKey()
         self.key.generate_key(OpenSSL.crypto.TYPE_RSA, 2048)
 
-        self.served = collections.defaultdict(set)
+        self.served = collections.defaultdict(set)  # type: ServedType
 
         # Stuff below is shared across threads (i.e. servers read
         # values, main thread writes). Due to the nature of CPython's
         # GIL, the operations are safe, c.f.
         # https://docs.python.org/2/faq/library.html#what-kinds-of-global-value-mutation-are-thread-safe
-        self.certs = {}
-        self.http_01_resources = set()
+        self.certs = {}  # type: Dict[bytes, Tuple[OpenSSL.crypto.PKey, OpenSSL.crypto.X509]]
+        self.http_01_resources = set() \
+        # type: Set[acme_standalone.HTTP01RequestHandler.HTTP01Resource]
 
         self.servers = ServerManager(self.certs, self.http_01_resources)
 
@@ -184,7 +207,7 @@ class Authenticator(common.Plugin):
     def add_parser_arguments(cls, add):
         add("supported-challenges",
             help=argparse.SUPPRESS,
-            type=supported_challenges_validator,
+            action=SupportedChallengesAction,
             default=",".join(chall.typ for chall in SUPPORTED_CHALLENGES))
 
     @property
@@ -219,46 +242,49 @@ class Authenticator(common.Plugin):
 
     def _perform_single(self, achall):
         if isinstance(achall.chall, challenges.HTTP01):
-            server, response = self._perform_http_01(achall)
+            servers, response = self._perform_http_01(achall)
         else:  # tls-sni-01
-            server, response = self._perform_tls_sni_01(achall)
-        self.served[server].add(achall)
+            servers, response = self._perform_tls_sni_01(achall)
+        self.served[servers].add(achall)
         return response
 
     def _perform_http_01(self, achall):
-        server = self.servers.run(self.config.http01_port, challenges.HTTP01)
+        port = self.config.http01_port
+        addr = self.config.http01_address
+        servers = self.servers.run(port, challenges.HTTP01, listenaddr=addr)
         response, validation = achall.response_and_validation()
         resource = acme_standalone.HTTP01RequestHandler.HTTP01Resource(
             chall=achall.chall, response=response, validation=validation)
         self.http_01_resources.add(resource)
-        return server, response
+        return servers, response
 
     def _perform_tls_sni_01(self, achall):
         port = self.config.tls_sni_01_port
-        server = self.servers.run(port, challenges.TLSSNI01)
+        addr = self.config.tls_sni_01_address
+        servers = self.servers.run(port, challenges.TLSSNI01, listenaddr=addr)
         response, (cert, _) = achall.response_and_validation(cert_key=self.key)
         self.certs[response.z_domain] = (self.key, cert)
-        return server, response
+        return servers, response
 
     def cleanup(self, achalls):  # pylint: disable=missing-docstring
-        # reduce self.served and close servers if none challenges are served
-        for server, server_achalls in self.served.items():
+        # reduce self.served and close servers if no challenges are served
+        for unused_servers, server_achalls in self.served.items():
             for achall in achalls:
                 if achall in server_achalls:
                     server_achalls.remove(achall)
-        for port, server in six.iteritems(self.servers.running()):
-            if not self.served[server]:
+        for port, servers in six.iteritems(self.servers.running()):
+            if not self.served[servers]:
                 self.servers.stop(port)
 
 
 def _handle_perform_error(error):
-    if error.socket_error.errno == socket.errno.EACCES:
+    if error.socket_error.errno == socket_errors.EACCES:
         raise errors.PluginError(
             "Could not bind TCP port {0} because you don't have "
             "the appropriate permissions (for example, you "
             "aren't running this program as "
             "root).".format(error.port))
-    elif error.socket_error.errno == socket.errno.EADDRINUSE:
+    elif error.socket_error.errno == socket_errors.EADDRINUSE:
         display = zope.component.getUtility(interfaces.IDisplay)
         msg = (
             "Could not bind TCP port {0} because it is already in "
